/*
** Module   : ClientBuilderOptions.java
** Abstract : store the parameters read from directory.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 MAG 20131128 First version.
** 002 MAG 20140128 Add WebSocket timeout.
** 003 MAG 20140204 Add Watchdog timeout.
** 004 CA  20140206 Refactored to make it unaware of the type of the P2J client being spawned.
*/

package com.goldencode.p2j.main;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.util.*;

/**
 * The parameters stored inside this structure are read from directory and are used as default 
 * parameters to create and spawn a P2J client process.
 */
public abstract class ClientBuilderOptions
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ClientBuilderOptions.class.getName());
   
   /** Directory node name */
   private static final String DIRECTORY_NODE_ID = "clientConfig";

   /** Secured flag */
   private boolean secure = true;
   
   /** Command */
   private String command = "java";

   /** User working directory */
   private String workingDirectory = "./";

   /** Class path */
   private String classPath;
   
   /** Native library path */
   private String libPath;

   /** Spawner */
   private String spawner;
   
   /** Client configuration file */
   private String configFile;

   /** JVM arguments */
   private List<String> jvmArguments = new LinkedList<String>();

   /** P2J client application arguments */
   private List<String> appArguments = new LinkedList<String>();

   /** P2J client bootstrap config */
   private List<String> bootstrapConfig = new LinkedList<String>();
   
   /** Main class, must have a main method */
   private Class<?> mainClass = ClientDriver.class;

   /**
    * Initialize. Read all configuration parameters from directory.
    */
   protected ClientBuilderOptions()
   {
      // Get default path
      String defaultPath = getDefaultPath();

      // Set default values using default path
      libPath = defaultPath;
      classPath = defaultPath  + "p2j.jar";
      spawner = defaultPath + "spawn";
      
      // Secure
      secure = getNode(DIRECTORY_NODE_ID, "secure", true);
      
      // Command
      command = quotedString(getNode(DIRECTORY_NODE_ID, "command", command));
      // Class path
      classPath = quotedString(getNode(DIRECTORY_NODE_ID, "classpath", classPath));
      // Native library path
      libPath = quotedString(getNode(DIRECTORY_NODE_ID, "libPath", libPath));
      
      // Spawner
      spawner = quotedString(getNode(DIRECTORY_NODE_ID, "spawner", spawner)); 
      
      // Working directory
      workingDirectory = quotedString(getNode(DIRECTORY_NODE_ID, "workingDir", workingDirectory)); 
      
      // Configuration file
      configFile = quotedString(getNode(DIRECTORY_NODE_ID, "configFile", configFile));
      
      // JVM arguments
      String jvm = getNode(DIRECTORY_NODE_ID, "jvmArgs",
                           "-Xmx512m -XX:MaxPermSize=64m -Djava.awt.headless=true");
      jvmArguments.addAll(Arrays.asList(jvm.split(" ")));
      
      // application arguments
      String appArgs = getNode(DIRECTORY_NODE_ID, "appArgs", "");
      appArguments.addAll(Arrays.asList(appArgs.split(" ")));
      
      // bootstrap config overrides
      String bootstrapCfg = getNode(DIRECTORY_NODE_ID, "cfgOverrides", "");
      bootstrapConfig.addAll(Arrays.asList(bootstrapCfg.split(" ")));
   }

   /**
    * When the string contains space characters a quoted string is build.
    * On Windows OS spaces could be used in file path, file name, font name and other
    * strings. In order to use such kind of strings in API calls the string should be
    * transformed in quoted form. 
    * 
    * @param   text  Text to search for spaces.
    * 
    * @return  If no spaces found return the initial string otherwise return a quoted string. 
    */
   protected static String quotedString(String text)
   {
      if (text == null)
      {
         return null;
      }
      
      if (text.startsWith("\"") && text.endsWith("\""))
      {
         return text;
      }
      return (text.indexOf(" ") > -1) ? "\\\"" + text +"\\\"" : text;
   }

   /**
    * Get the authentication mode for this spawner.
    * 
    * @return   {@code true} indicating that the user's password is required for authentication.
    */
   protected abstract boolean isPasswordAuthentication();
   
   /**
    * Get the security mode.
    * 
    * @return  <code>true</code> use a secured connection <code>false</code> otherwise.
    */
   public boolean isSecure()
   {
      return secure;
   }

   /**
    * Get command
    * 
    * @return  Command name string.
    */
   public String getCommand()
   {
      return command;
   }
   
   /**
    * Get classpath
    * 
    * @return  Class path string.
    */
   public String getClassPath()
   {
      return classPath;
   }

   /**
    * Get native library path.
    * 
    * @return  Native library path string.
    */
   public String getLibPath()
   {
      return libPath;
   }

   /**
    * Get user working directory.
    * 
    * @return  User working directory.
    */
   public String getWorkingDirectory()
   {
      return workingDirectory;
   }

   /**
    * Get configuration file name.
    * 
    * @return  Configuration file name.
    */
   public String getConfigFile()
   {
      return configFile;
   }
   
   /**
    * Get JVM arguments
    * 
    * @return  A List with JVM arguments.
    */
   public List<String> getJvmArguments()
   {
      return jvmArguments;
   }
   
   /**
    * Get P2J client's application arguments.
    * 
    * @return  A List with application arguments.
    */
   public List<String> getAppArguments()
   {
      return appArguments;
   }
   
   /**
    * Get P2J client's bootstrap config.
    * 
    * @return  A List with bootstrap config settings.
    */
   public List<String> getBootstrapConfig()
   {
      return bootstrapConfig;
   }

   /**
    * Get spawner command.
    * 
    * @return  The spawner command;
    */
   public String getSpawner()
   {
      return spawner;
   }

   /**
    * Get the Java process main class.
    * 
    * @return  The main class.
    */
   public Class<?> getMainClass()
   {
      return mainClass;
   }

   /**
    * Get a string value from a directory node.
    *  
    * @param   parent
    *          The name of the parent node.
    * @param   node
    *          Node name.
    * @param   defa
    *          Default value if node not found.
    *          
    * @return  Node value if exists or the default value otherwise.
    */
   protected String getNode(String parent, String node, String defa)
   {
      return Utils.getDirectoryNodeString(null, parent + "/" + node, defa, false);
   }

   /**
    * Get an integer value from a directory node.
    * 
    * @param   parent
    *          The name of the parent node.
    * @param   node
    *          Node name.
    * @param   defa
    *          Default value if node not found.
    *          
    * @return  Node value if exists or the default value otherwise.
    */
   protected int getNode(String parent, String node, int defa)
   {
      return Utils.getDirectoryNodeInt(null, parent + "/" + node, defa, false);
   }

   /**
    * Get a boolean value from a directory node.
    *  
    * @param   parent
    *          The name of the parent node.
    * @param   node
    *          Node name.
    * @param   defa
    *          Default value if node not found.
    *          
    * @return  Node value if exists or the default value otherwise.
    */
   protected boolean getNode(String parent, String node, boolean defa)
   {
      return Utils.getDirectoryNodeBoolean(null, parent + "/" + node, defa, false);
   }

   /**
    * Get default path as the p2j.jar file location.
    * 
    * @return  Path to p2j.jar file.
    */
   private String getDefaultPath()
   {
      String message = "Could not determine the location of the p2j.jar file.";
      /** Default value */
      String defaultPath = "./"; 
      try
      {
         URL url = mainClass.getProtectionDomain().getCodeSource().getLocation();
         String classPath = url.toURI().getPath();
         File jarFile = new File(classPath);
         // If the source is in a folder and not jar, than this can cause problems
         if (!jarFile.exists())
         {
            LOG.logp(Level.WARNING,
                     "ClientBuilderOptions.getDefaultPath()",
                     "",
                     LogHelper.generate(message));
         }
         else
         {
            // Default path
            defaultPath = jarFile.getParentFile().getPath();
            if (!defaultPath.endsWith(File.separator))
            {
               defaultPath += File.separator;
            }
         }
      } 
      catch (URISyntaxException e)
      {
         LOG.logp(Level.WARNING,
                  "ClientBuilderOptions.getDefaultPath()",
                  "",
                  LogHelper.generate(message), 
                  e);
      }
      
      return defaultPath;
   }
}
