/*
** Module   : ClientSpawner.java
** Abstract : define a generic client Spawner.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 MAG 20131119 First version.
** 002 CA  20140206 Refactored to make it unaware of the type of the P2J client being spawned.
*/

package com.goldencode.p2j.main;

import java.util.concurrent.*;
import java.util.logging.*;

import com.goldencode.p2j.util.*;

/**
 * This defines a way of spawning new P2J client processes.  Temporary credentials are sent to
 * the P2J client side, which, if used, will establish a P2J session with the P2J server and do
 * some custom work (depending on the implementation).
 * <p>
 * Actual implementations of this class will need to provide the following info:
 * <ul>
 *    <li>A {@link ClientBuilder} instance, to spawn the P2J client process.</li>
 *    <li>A {@link ClientBuilderParameters} instance, to configure the P2J client.</li>
 *    <li>Implementation for the {@link SpawnerListener#getServerData()}, to send custom data
 *        back to the P2J client.</li>
 *    <li>Implementation for the {@link SpawnerListener#getTemporaryClient()}, with the work
 *        needed to be done after authenticating with the temporary credentials.</li>
 * </ul>
 */
public abstract class ClientSpawner
implements SpawnerListener
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ClientSpawner.class.getName());
   
   /** Timeout */
   private static final int TIME_OUT = 30;
   
   /** The {@link ClientBuilder} used to spawn the P2J client process. */
   protected ClientBuilder cb;
   
   /** The configuration for this spawner. */
   protected final ClientBuilderParameters cfg;

   /** Spawner exit code */
   private int exitCode = 0;

   /** Synchronize */
   private CountDownLatch latch;

   /** Flag indicating the P2J client has been started. */
   private volatile boolean ready = false;

   /**
    * Configure a new client spawner.
    * 
    * @param    cfg
    *           The spawner configuration.
    */
   public ClientSpawner(ClientBuilderParameters cfg)
   {
      this.cfg = cfg;
   }
   
   /**
    * Get a {@link ClientBuilder} used to spawn P2J clients.
    * 
    * @param    cfg
    *           The P2J client configuration.
    *           
    * @return   See above.
    */
   protected abstract ClientBuilder getBuilder(ClientBuilderParameters cfg);

   /**
    * Launch a P2J client.
    * 
    * @param   userName
    *          User name.
    * @param   password
    *          User password.
    */
   public void spawn(String userName, String password)
   {
      if (userName == null)
      {
         throw new NullPointerException("The username must be provided!");
      }
      
      // Get a temporary account from pool.
      TemporaryAccount tmpAcc = TemporaryAccount.open();
      if (tmpAcc == null)
      {
         String message =
            "Temporary Account Pool is empty! Please call your system administrator.";

         LOG.logp(Level.WARNING, "ClientSpawner.spawn()", "", LogHelper.generate(message));
         
         throw new RuntimeException(message);
      }
      
      cfg.setUserName(userName);
      cfg.setPassword(password);
      
      cb = getBuilder(cfg);
      
      // register listener
      SpawnerImpl spawner = (SpawnerImpl) SpawnerImpl.getInstance();
      spawner.addListener(cfg.getUuid(), this);

      // spawn
      Process shell = null;
      ready = false;
      try
      {
         latch = new CountDownLatch(1);
         
         // spawn 
         shell = cb.start(tmpAcc.getEnvironmentMap());
         
         // wait for notification
         latch.await(TIME_OUT, TimeUnit.SECONDS);
         
         shell.waitFor();
         exitCode = shell.exitValue();
      }
      catch (Exception e)
      {
         LOG.logp(Level.SEVERE, 
                  "ClientSpawner.spawn()", 
                  "",
                  LogHelper.generate("Cannot spawn the process."),
                  e);
      }
      finally
      {
         spawner.removeListener(cfg.getUuid());
         tmpAcc.close();
         
         // cleanup the command
         ClientBuilder.removeCommand(cfg.getUuid());
      }

      if (!ready)
      {
         shell.destroy();
      }
   }

   /**
    * Return spawner exit code
    * 
    * @return  Spawn exit value.
    */
   public int getExitCode()
   {
      return exitCode;
   }
   
   /**
    * Notify when the client has started.
    * 
    * @param    data
    *           Custom data sent by the P2J client back to the server.
    */
   @Override
   public void clientIsReady(Object data)
   {
      this.ready = true;
      latch.countDown();
   }
}
