/*
** Module   : Job.java
** Abstract : POJO defining a scheduled job.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140206 First version.
*/

package com.goldencode.p2j.main;

import java.util.*;

/**
 * Definition of a {@link Scheduler} job.
 * <p>
 * For each job, the following can be defined:
 * <ul>
 *    <li>{@link #name} - the job name; must be unique and is case insensitive.</li>
 *    <li>{@link #type} - the kind of the job being started:
 *       <p> {@link JobType#PROCESS} - start a new P2J process.
 *       <p>-{@link JobType#CLASS} - execute a certain java class which implements 
 *           {@link Runnable}.
 *       <p>-{@link JobType#METHOD} - execute a certain java method.
 *    </li>
 *    <li>{@link #target}
 *       <p>- in case of {@link JobType#PROCESS}, is the subject ID of a P2J process.
 *       <p>- in case of {@link JobType#CLASS}, is the qualified name of a Java class which  
 *            implements the {@link Runnable} interface 
 *       <p>- in case of {@link JobType#METHOD}, is the qualified name of a Java class followed by 
 *            the name of a no-arguments method, i.e. {@code foo.bar.Something.execute}
 *       </p>
 *    </li>
 *    <li>{@link #enabled} - flag indicating if this job is enabled or not.</li>
 *    <li>{@link #mode} - the job mode. One of the following values:
 *       <p>- {@link JobMode#NOW} - for jobs being executed immediately
 *       <p>- {@link JobMode#ONE_TIME} - for jobs being executed at a certain time/date,
 *           a single time.
 *       <p>- {@link JobMode#RECURRENT} - for recurrent jobs executed at a certain time/date.
 *    </li>
 *    <li>The {@link #minute}, {@link #hour}, {@link #dayOfMonth}, {@link #dayOfMonth} and 
 *        {@link #month} need to be specified only for {@link JobMode#RECURRENT} or 
 *        {@link JobMode#ONE_TIME} jobs.  The following rules apply when scheduling a job:
 *        <p>- the {@link #minute} is mandatory, one of [0-59] values.
 *        <p>- the {@link #hour} is optional, one of [0-23] values.
 *        <p>- {@link #dayOfMonth} is optional and is one of the [1-31] values, following the
 *             {@link Calendar#DAY_OF_MONTH} specification; value 32 represens the "end of the
 *             month day".
 *        <p>- {@link #dayOfWeek} is optional and is one of the [1-7] values, following the
 *             {@link Calendar#DAY_OF_WEEK} specification. Sunday is value 1.
 *        <p>- {@link #month} is optional and is one of [0-11] values, following
 *             {@link Calendar#MONTH} specification.  If specified, one of the {@link #dayOfWeek}
 *             or {@link #dayOfMonth} must be specified too.
 *        <p>- only one of {@link #dayOfMonth} and {@link #dayOfWeek} can be specified.
 *        <p>- if {@link #dayOfMonth} is specified and {@link #month} is specified, it must be
 *           a valid day for the specified {@link #month}.
 *        <p>- if {@link #dayOfMonth} is specified and {@link #month} is not specified, the 
 *           scheduler will execute the job only in the months which have this day; I.e. setting
 *           this to 31 will mean the job will be executed only in months which have 31 days.
 *           Use value 32 to execute the job on the last day of the month.
 *     </li>
 * </ul>
 */
public class Job
{
   /** The name of this job. */
   private final String name;

   /** The type of this job, one of the {@link JobType} values. */
   private JobType type;

   /** The target of this job. */
   private String target;
   
   /** Flag indicating if this job is enabled. */
   private boolean enabled;
   
   /** The execution mode of this job, one of the {@link JobMode} values. */
   private JobMode mode;

   /** The minute when the job needs to be executed. */
   private Integer minute;
   
   /** The hour when the job needs to be executed. */
   private Integer hour;
   
   /** The day the month when the job needs to be executed. */
   private Integer dayOfMonth;
   
   /** The day of the week when the job needs to be executed. */
   private Integer dayOfWeek;
   
   /** The month when the job needs to be executed. */
   private Integer month;
   
   /**
    * Create a new job.
    * 
    * @param    name
    *           The name of this job.
    * @param    type
    *           target
    * @param    target
    *           The target of this job.
    * @param    enabled
    *           Flag indicating if this job is enabled.
    * @param    mode
    *           The execution mode of this job, one of the {@link JobMode} values.
    */
   public Job(String name, JobType type, String target, boolean enabled, JobMode mode)
   {
      this.name = name;
      this.type = type;
      this.target = target;
      this.enabled = enabled;
      this.mode = mode;
   }

   /**
    * Get the job's {@link #name}.
    * 
    * @return   See above.
    */
   public String getName()
   {
      return name;
   }

   /**
    * Get the job's {@link #type}.
    * 
    * @return   See above.
    */
   public JobType getType()
   {
      return type;
   }

   /**
    * Get the job's {@link #target}.
    * 
    * @return   See above.
    */
   public String getTarget()
   {
      return target;
   }

   /**
    * Get the job's {@link #enabled} state.
    * 
    * @return   See above.
    */
   public boolean isEnabled()
   {
      return enabled;
   }

   /**
    * Get the job's {@link #mode}.
    * 
    * @return   See above.
    */
   public JobMode getMode()
   {
      return mode;
   }

   /**
    * Get the job's {@link #minute}.
    * 
    * @return   See above.
    */
   public Integer getMinute()
   {
      return minute;
   }

   /**
    * Set the job's execution {@link #minute}.
    * 
    * @param    minute
    *           The minute, {@code null} or a [0 - 59] value.
    */
   public void setMinute(Integer minute)
   {
      this.minute = minute;
   }

   /**
    * Get the job's {@link #hour}.
    * 
    * @return   See above.
    */
   public Integer getHour()
   {
      return hour;
   }

   /**
    * Set the job's execution {@link #hour}.
    * 
    * @param    hour
    *           The hour, {@code null} or a [0 - 23] value.
    */
   public void setHour(Integer hour)
   {
      this.hour = hour;
   }

   /**
    * Get the job's {@link #dayOfMonth}.
    * 
    * @return   See above.
    */
   public Integer getDayOfMonth()
   {
      return dayOfMonth;
   }

   /**
    * Set the job's execution {@link #dayOfMonth}.
    * 
    * @param    dayOfMonth
    *           The day of the month, {@code null} or a [1 - 31] value, following the 
    *           {@link Calendar#DAY_OF_MONTH} specification.  Value 32 is used to represent the
    *           last day of the month.
    */
   public void setDayOfMonth(Integer dayOfMonth)
   {
      this.dayOfMonth = dayOfMonth;
   }

   /**
    * Get the job's {@link #dayOfWeek}.
    * 
    * @return   See above.
    */
   public Integer getDayOfWeek()
   {
      return dayOfWeek;
   }

   /**
    * Set the job's execution {@link #dayOfWeek}.
    * 
    * @param    dayOfWeek
    *           The day of the week, {@code null} or a [1 - 7] value where 1 is for Sunday, 
    *           following the {@link Calendar#DAY_OF_WEEK} specification.
    */
   public void setDayOfWeek(Integer dayOfWeek)
   {
      this.dayOfWeek = dayOfWeek;
   }

   /**
    * Get the job's {@link #month}.
    * 
    * @return   See above.
    */
   public Integer getMonth()
   {
      return month;
   }

   /**
    * Set the job's execution {@link #month}.
    * 
    * @param    month
    *           The month, {@code null} or a [0 - 11] value following the {@link Calendar#MONTH} 
    *           specification.
    */
   public void setMonth(Integer month)
   {
      this.month = month;
   }
   
   /**
    * Get the string representation of this job, using the 
    * {@code "<name> [<type>:<mode>] - <target>"} format.
    */
   @Override
   public String toString()
   {
      return String.format("%s [%s:%s] - %s", this.name, this.type, this.mode, this.target);
   }
}
