/*
** Module   : ProcessClientSpawner.java
** Abstract : define the P2J Process Client Spawner.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20140206 First version.
*/

package com.goldencode.p2j.main;

import java.util.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.client.chui.driver.web.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.util.*;

/**
 * Class used to spawn a new P2J client which will authenticate as a P2J process.
 * <p>
 * The spawned process first try to authenticate the user against OS. On success a new process 
 * (client) is spawned. The spawned process will get the secure credentials from the P2J server,
 * start the client and notify the P2J server that the P2J process will start.
 * <p>
 * This notification must not be treated as the P2J process authenticated successfully; it needs
 * to be interpreted as a signatl that the temporary connection has finished successfully.
 */
public class ProcessClientSpawner
extends ClientSpawner
{
   /** The P2J process ID being spawned by this instance.*/
   private String process;

   /**
    * Create a new spawner for P2J processes.
    * 
    * @param    process
    *           The P2J process ID.
    */
   public ProcessClientSpawner(String process)
   {
      super(new ClientBuilderParameters(new ProcessBuilderOptions(process)));
      
      this.process = process;
   }

   /**
    * Launch the specified batch process.
    * <p>
    * If the P2J process is linked with an appserver, it will delegate the launching to
    * {@link AppServerLauncher#launch}.  Else, it will use the {@link ProcessClientSpawner 
    * spawner} infrastructure to start a P2J client.
    * 
    * @param    batchProcess
    *           The name of the batch process to be launched.
    */
   public static void launch(String batchProcess)
   {
      SecurityManager sm = SecurityManager.getInstance();

      String appserver = sm.getAppserverForProcess(batchProcess);
      if (appserver != null)
      {
         // appserver's are special, so treat the now
         AppServerLauncher.launch(appserver);
      }
      else
      {
         // normal process, launch it
         ProcessClientSpawner spawner = new ProcessClientSpawner(batchProcess);

         // get the user name from the directory
         spawner.spawn(sm.getSystemUser(batchProcess), null);
      }
   }
   
   /**
    * Get a map containing the P2J process secure credentials: the serialized process keystore and
    * trust store and other details, like target aliases in the keystore and truststore. 
    */
   @Override
   public Object getServerData()
   {
      // return a map with the process defaults
      Map<String, Object> defaults = new HashMap<>();
      
      ProcessBuilderOptions opts = (ProcessBuilderOptions) cb.config.options;
      
      if (opts.getProcessAlias() != null)
         defaults.put("security:keystore:processalias", opts.getProcessAlias());
      if (opts.getTruststoreAlias() != null)
         defaults.put("security:truststore:alias", opts.getTruststoreAlias());
      if (opts.getKeyentryPassword() != null)
         defaults.put("access:password:keystore", opts.getKeystorePassword());
      if (opts.getKeyentryPassword() != null)
         defaults.put("access:password:keyentry", opts.getKeyentryPassword());
      if (opts.getTruststorePassword() != null)
         defaults.put("access:password:truststore", opts.getTruststorePassword());
      if (opts.isValidateCertificate())
         defaults.put("security:certificate:validate", opts.isValidateCertificate());
      if (opts.getProcessKeystore() != null)
         defaults.put("security:keystore:bytes", opts.getProcessKeystore());
      if (opts.getTruststore() != null)
         defaults.put("security:truststore:bytes", opts.getTruststore());
      
      return defaults;
   }

   /**
    * Get a {@link ProcessClientBuilder} used to spawn P2J clients for the target {@link #process}.
    * 
    * @param    cfg
    *           The P2J web client configuration.
    *           
    * @return   See above.
    */
   @Override
   protected ClientBuilder getBuilder(ClientBuilderParameters cfg)
   {
      return new ProcessClientBuilder(cfg);
   }

   /**
    * Get a {@link TemporaryClient} worker which will do the work after authenticating on the P2J
    * server using the temporary credentials.
    * 
    * @return   See above.
    */
   @Override
   public TemporaryClient getTemporaryClient()
   {
      return new TemporaryClientTask();
   }
   
   /**
    * On P2J client side, it will do custom work after authenticating using the temporary 
    * credentials.
    * <p>
    * For process clients case, it will retrieve from the P2J server the secure credentials
    * and save them in the {@link BootstrapConfig configuration} used to start the P2J process.
    */
   public static class TemporaryClientTask
   implements TemporaryClient
   {
      /**
       * On P2J client side, it will do custom work after authenticating using the temporary 
       * credentials.
       * <p>
       * For web chui clients case, it will create a {@link ChuiWebDriver} and start the web 
       * embedded server which will serve this client's requests.
       * <p>
       * Any custom configuration already existing in the {@code config} parameter will not
       * be overwritten.
       * 
       * @param    spawner
       *           The spawner remote access interface. 
       * @param    uuid
       *           Remote client identifier uuid.
       * @param    config
       *           The configuration to use for setup of the client and for the server connection.
       *
       * @return   Always {@code null}.
       * 
       * @throws   RuntimeException
       *           If there are problems accessing the secure credentials.
       */
      @Override
      public ScreenDriver doWork(Spawner spawner, String uuid, BootstrapConfig config)
      {
         // get the process key store and trust stores from the server
         @SuppressWarnings("unchecked")
         Map<String, Object> defaults = (Map<String, Object>) spawner.exportData(uuid);

         String  alias       = (String)  defaults.get("security:keystore:processalias");
         String  tsAlias     = (String)  defaults.get("security:truststore:alias");
         String  ksPass      = (String)  defaults.get("access:password:keystore");
         String  kePass      = (String)  defaults.get("access:password:keyentry");
         String  tsPass      = (String)  defaults.get("access:password:truststore");
         Boolean validate    = (Boolean) defaults.get("security:certificate:validate");
         byte[]  keystore    = (byte[])  defaults.get("security:keystore:bytes");
         byte[]  truststore  = (byte[])  defaults.get("security:truststore:bytes");

         try
         {
            // save them at the bootstrap config. any client-side config has priority
            if (alias != null &&
                config.getConfigItem("security", "keystore", "processalias") == null)
            {
               config.setConfigItem("security", "keystore", "processalias", alias);
            }
            
            if (tsAlias != null &&
                config.getConfigItem("security", "truststore", "alias") == null)
            {
               config.setConfigItem("security", "truststore", "alias", tsAlias);
            }
            
            if (ksPass != null && 
                config.getConfigItem("access", "password", "keystore") == null)
            {
               config.setConfigItem("access", "password", "keystore", ksPass);
            }
            
            if (kePass != null &&
                config.getConfigItem("access", "password", "keyentry") == null)
            {
               config.setConfigItem("access", "password", "keyentry", kePass);
            }
            
            if (tsPass != null &&
               config.getConfigItem("access", "password", "truststore") == null)
            {
               config.setConfigItem("access", "password", "truststore", tsPass);
            }
            
            if (validate != null &&
                config.getConfigItem("security", "certificate", "validate") == null)
            {
               config.setConfigItem("security", "certificate", "validate", validate.toString());
            }
            
            if (keystore != null &&
                config.getConfigItem("security", "keystore", "filename") == null)
            {
               config.setConfigItem("security", "keystore", "bytes",
                                    Base64.byteArrayToBase64(keystore));
            }

            if (truststore != null &&
               config.getConfigItem("security", "truststore", "filename") == null)
            {
               config.setConfigItem("security", "truststore", "bytes",
                                    Base64.byteArrayToBase64(truststore));
            }
            
            // invalidate the security transport...
            config.setConfigItem("security", "transport", "refresh", "true");

            spawner.clientIsReady(uuid, null);
         }
         catch (ConfigurationException e)
         {
            throw new RuntimeException(e);
         }
         
         return null;
      }
  }
}
