/*
** Module   : Scheduler.java
** Abstract : cron-like scheduler for P2j processes.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140206 First version.
*/

package com.goldencode.p2j.main;

import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.*;

/**
 * Implements a cron-like scheduler, which can run P2J processes or arbitrary Java classes or
 * methods.
 * <p>
 * The scheduler is configured in the {@code scheduler} container, which resides as a child of the
 * {@code server/<server-id>} or the {@code server/default} nodes.
 * <p>
 * The {@code scheduler} node has a children nodes of class {@code job}. For each {@code job} node,
 * the node's name represents the job name (which must be unique accross all jobs).  Also, 
 * {@code node-attribute} nodes can be provided, with the following name:
 * <ul>
 *    <li>{@code type}, mandatory, case insensitive: string representation of the job type,  
 *        following the description at the {@link JobType} enum.</li>
 *    <li>{@code target}, mandatory: the job target, which can be a process name, a java class or
 *        a java method (see javadoc for {@link Job} class for more details.</li>
 *    <li>{@code enabled}, mandatory: boolean value flag indicating if this job is enabled.</li>
 *    <li>{@code mode}, mandatory: string representation of the job mode, following the 
 *         description at the {@link JobMode} enum.</li>
 *    <li>{@code minute}: the scheduled minute.</li>
 *    <li>{@code hour}: the scheduled hour.</li>
 *    <li>{@code day-of-week}: the scheduled day of the week.</li>
 *    <li>{@code day-of-month}: the scheduled day of the month.</li>
 *    <li>{@code month}: the scheduled month.</li>
 * </ul>
 * <p>
 * An example of a job starting an appserver at server startup:
 * <pre>
 * {@code
 *   <node class="container" name="scheduler">
 *     <node class="job" name="start_appserver">
 *       <node-attribute name="type" value="process"/>
 *       <node-attribute name="target" value="appserver-name"/>
 *       <node-attribute name="enabled" value="TRUE"/>
 *       <node-attribute name="mode" value="now"/>
 *     </node>
 *   </node>
 * }
 * </pre>
 * <p>
 * WARNING: when searching for the scheduler node, if there are both 
 * {@code server/<server-id>/scheduler} and {@code server/default/scheduler}, only the server's
 * scheduler node will be used.  Currently, there is no way of merging the default and server's
 * scheduler. 
 * <p>
 * When scheduling jobs for execution, each job must run under the server's context.  This is not
 * possible using the {@link #timer}'s threads.  To solve this, the approach is to use the 
 * {@link #timer} as a proxy which sends the jobs to the {@link JobProcessor}, which runs in its
 * own thread, under the server's context.  When the {@link JobProcessor} receives a new job,
 * it will create a dedicated thread (with the server's context) and use that thread to execute 
 * the job.
 * 
 * @see   Job
 *        For more details about how jobs are scheduled.
 */
public final class Scheduler
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(Scheduler.class);
   
   /** A map identifying all the available jobs. */
   private static Map<String, Job> jobs = null;
   
   /** A registry with the enabled jobs, scheduled to be ran. */
   private static Map<String, JobTask> jobRegistry = new HashMap<>();
   
   /** Job launcher timer.  Will execute each job at its scheduled time. */
   private static Timer timer = new Timer("Job launcher", true);
   
   /**
    * This instance will run on a dedicated thread, under the server's context, as job launching
    * requires access to the directory.
    */
   private static final JobProcessor processor = new JobProcessor();
   
   /**
    * Do not allow instances of this class.
    */
   private Scheduler()
   {
      // do not allow instances of this class
   }
   
   /**
    * Check if the specified job is scheduled for execution.
    * 
    * @param    jobName
    *           The job name.
    *           
    * @return   See above.
    */
   public static boolean isScheduled(String jobName)
   {
      jobName = jobName.toLowerCase();
      
      synchronized (jobRegistry)
      {
         return jobRegistry.containsKey(jobName);
      }
   }
   
   /**
    * Get the next execution time for the given job.
    * 
    * @param    jobName
    *           The job name.
    *           
    * @return   The next execution time or {@code -1}, if not scheduled.
    */
   public static long scheduledExecutionTime(String jobName)
   {
      jobName = jobName.toLowerCase();
      
      synchronized (jobRegistry)
      {
         if (jobRegistry.containsKey(jobName))
         {
            return jobRegistry.get(jobName).scheduledExecutionTime();
         }
      }
      
      return -1;
   }
   
   /**
    * Remove the given job from the {@link #jobs global job registry}.  Active jobs (which are
    * scheduled for execution) can not be removed.  Use {@link #cancelJob} to cancel it first. 
    * 
    * @param    jobName
    *           The job name.
    *           
    * @return   An {@link Job} instance with the removed job details or {@code null} if the job
    *           is scheduled for execution or it does not exist. 
    */
   public static Job removeJob(String jobName)
   {
      jobName = jobName.toLowerCase();
    
      synchronized (jobRegistry)
      {
         if (jobRegistry.containsKey(jobName))
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               String msg = "Can not remove job - there are pending tasks! Cancel the job first.";
               LOG.logp(Level.WARNING, "Scheduler", "removeJob", LogHelper.generate(msg));
               return null;
            }
         }
      }
      
      synchronized (jobs)
      {
         return jobs.remove(jobName);
      }
   }

   /**
    * Get the {@link Job} with the given name.
    * 
    * @param    jobName
    *           The job name.
    *           
    * @return   The {@link Job} instance or {@code null} if it does not exist.
    */
   public static Job getJob(String jobName)
   {
      jobName = jobName.toLowerCase();
      
      synchronized (jobs)
      {
         return jobs.get(jobName);
      }
   }
   
   /**
    * Add a new {@link Job} to the {@link #jobs global registry}.
    *  
    * @param    job
    *           The job name.  Must not already be in use for another job.
    *           
    * @return   <code>true</code> if the job was added. <code>false</code> if there is already
    *           a job with the same name,
    */
   public static boolean addJob(Job job)
   {
      String jobName = job.getName().toLowerCase();
      
      synchronized (jobs)
      {
         if (jobs.containsKey(jobName))
         {
            return false;
         }
         
         jobs.put(jobName, job);
      }
      
      return true;
   }
   
   /**
    * Cancel the job with the given name.
    * 
    * @param    jobName
    *           The job name.
    *           
    * @return   <code>true</code> if the job was scheduled for execution and it was canceled.
    */
   public static boolean cancelJob(String jobName)
   {
      jobName = jobName.toLowerCase();
      
      synchronized (jobRegistry)
      {
         JobTask task = jobRegistry.get(jobName);

         if (task != null)
         {
            task.cancel();

            // remove the reference from the timer
            timer.purge();
         }
         
         return (task != null);
      }
   }
   
   /**
    * Schedule a job for execution.
    * <p>
    * First, this will cancel any pending schedules and schedule it again, using the latest 
    * {@link Job} configuration.
    * <p>
    * Second, it will create a new {@link JobTask} which will be registered with the 
    * {@link #timer}.  At appropriate time, the {@link #timer} will pick up the {@link JobTask},
    * use {@link #getRunner(Job)} to create a new {@link Runnable} instance for this job and
    * send it to the {@link JobProcessor} thread.
    * 
    * @param    jobName
    *           The job name.
    * @param    lastExecutionTime
    *           The last time the job was executed. {@code null} if this is the first time the
    *           job needs to run.
    *           
    * @return   {@code true} if the job was scheduled.
    */
   public static boolean scheduleJob(String jobName, Long lastExecutionTime)
   {
      jobName = jobName.toLowerCase();

      synchronized (jobs)
      {
         Job job = getJob(jobName);
         
         if (job == null)
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               String msg = String.format("Job '%s' is not registered with the scheduler!", 
                                          jobName);
               LOG.logp(Level.WARNING, "Scheduler", "scheduleJob", LogHelper.generate(msg));
            }

            return false;
         }
         
         synchronized (jobRegistry)
         {
            // make sure this is not already scheduled
            cancelJob(jobName);
            
            // schedule the job for a new run
            long next = getNextScheduleTime(job, lastExecutionTime);
            if (next < 0)
            {
               // can not schedule it
               return false;
            }
            
            JobTask runner = new JobTask(jobName);

            jobRegistry.put(jobName, runner);

            if (LOG.isLoggable(Level.FINE))
            {
               String msg = String.format("Scheduling job %s to execute at %tc.", jobName, next);
               LOG.logp(Level.WARNING, "Scheduler", "scheduleJob", LogHelper.generate(msg));
            }
            
            timer.schedule(runner, new Date(next));
         }
      }
      
      return true;
   }
   
   /**
    * Validate the given configuration and create a new job.
    * 
    * @param    jobName
    *           The name of this job.
    * @param    type
    *           The type of this job, one of the {@link JobType} values.
    * @param    target
    *           The target of this job.
    * @param    enabled
    *           Flag indicating if this job is enabled.
    * @param    mode
    *           The execution mode of this job, one of the {@link JobMode} values.
    * @param    minute
    *           The minute when the job needs to be executed.
    * @param    hour
    *           The hour when the job needs to be executed.
    * @param    dayOfWeek
    *           The day of the week when the job needs to be executed.
    * @param    dayOfMonth
    *           The day the month when the job needs to be executed.
    * @param    month
    *           The month when the job needs to be executed.
    * 
    * @return   A new job instance.
    * 
    * @throws   JobValidationException
    *           If the configuration can't be validated.  See the {@link Job} class javadoc for
    *           details about the validations being performed.
    */
   public static Job createJob(String  jobName,
                               String  type,
                               String  target,
                               boolean enabled,
                               String  mode,
                               Integer minute,
                               Integer hour,
                               Integer dayOfWeek,
                               Integer dayOfMonth,
                               Integer month)
   throws JobValidationException
   {
      if (type == null)
      {
         final String msg = "Type for job %s is missing: use 'process' or 'java'.";
         throw new JobValidationException(String.format(msg, jobName));
      }
      
      JobType jobType = JobType.resolveJobType(type);
      if (jobType == JobType.PROCESS)
      {
         SecurityManager sm = SecurityManager.getInstance();

         // only processes can be set to be scheduled
         if (!sm.isProcessDefined(target))
         {
            final String msg = "Scheduler: process '%s' does not exist for job %s!";
            throw new JobValidationException(String.format(msg, target, jobName));
         }
         
      }
      else if (jobType == JobType.CLASS)
      {
         String msg = null;

         try
         {
            Class<?> cls = Class.forName(target);
            int mods = cls.getModifiers();

            if (Modifier.isAbstract(mods) || Modifier.isInterface(mods))
            {
               msg = "Job '%s': class '%s' must be a non-abstract Java class!";
            }
            else if (Runnable.class.isAssignableFrom(cls))
            {
               // must implement runnable
               msg = "Job '%s': class '%s' must implement Runnable interface!";
            }
         }
         catch (ClassNotFoundException e)
         {
            msg = "Job '%s': class '%s' could not be found!";

            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.logp(Level.SEVERE, "Scheduler", "createJob",
                        String.format(msg, jobName, target), e);
            }
         }
         
         if (msg != null)
         {
            throw new JobValidationException(String.format(msg, jobName, target));
         }
      }
      else if (jobType == JobType.METHOD)
      {
         if (target.indexOf('.') == -1)
         {
            String msg = "Job '%s': target '%s' has no method name!";
            throw new JobValidationException(String.format(msg, jobName, target));
         }
         else
         {
            String msg = null;
            String clsName = target.substring(0, target.lastIndexOf('.'));
            String mthdName = target.substring(target.lastIndexOf('.') + 1);

            try
            {
               Class<?> cls = Class.forName(target);
               int mods = cls.getModifiers();
   
               if (Modifier.isAbstract(mods)  ||
                   Modifier.isInterface(mods) ||
                   !Modifier.isPublic(mods))
               {
                  msg = "Job '%s': class '%s' must be a non-abstract, public Java class!";
               }
               else if (!Runnable.class.isAssignableFrom(cls))
               {
                  // must implement runnable
                  msg = "Job '%s': class '%s' must implement Runnable interface!";
               }
               else
               {
                  // validate the method
                  Method mthd = cls.getMethod(mthdName);
                  
                  mods = mthd.getModifiers();
                  if (!Modifier.isPublic(mods))
                  {
                     msg = "Job '%s' class '%s': method '%s' must be public!";
                  }
               }
            }
            catch (ClassNotFoundException e)
            {
               msg = "Job '%s': class '%s' for method '%s' could not be found!";
   
               if (LOG.isLoggable(Level.SEVERE))
               {
                  LOG.logp(Level.SEVERE, "Scheduler", "createJob",
                           String.format(msg, jobName, clsName, mthdName), e);
               }
            }
            catch (NoSuchMethodException | SecurityException e)
            {
               msg = "Job '%s' class '%s': method '%s' could not be found!";
               
               if (LOG.isLoggable(Level.SEVERE))
               {
                  LOG.logp(Level.SEVERE, "Scheduler", "createJob",
                           String.format(msg, jobName, clsName, mthdName), e);
               }
            }

            if (msg != null)
            {
               throw new JobValidationException(String.format(msg, jobName, target));
            }
         }
      }
      else
      {
         String msg = "Invalid type '%s' for job %s (use 'process' or 'java').";
         throw new JobValidationException(String.format(msg, type, jobName));
      }
      
      if (mode == null)
      {
         String msg = "Mode for job %s is missing: use 'now', 'recurrent' or 'one-time'";
         throw new JobValidationException(String.format(msg, jobName));
      }
      
      JobMode jobMode = JobMode.resolveJobMode(mode);
      if (jobMode == null)
      {
         String msg = "Invalid mode '%s' for job %s (use 'now', 'recurrent' or 'one-time').";
         throw new JobValidationException(String.format(msg, mode, jobName));
      }
      
      // validate the scheduled time
      if (jobMode.isTimed())
      {
         // the minute is mandatory
         validValue(jobName, minute, true, "minute", 0, 59);
         // hour is optional
         validValue(jobName, hour, false, "hour", 0, 23);
         // day-of-week is optional if no month is specified
         validValue(jobName, dayOfWeek, (month != null), "day-of-week", 1, 7);
         // day-of-month is optional if no month is specified
         validValue(jobName, dayOfMonth, (month != null), "day-of-month", 1, 32);

         if (dayOfWeek != null && dayOfMonth != null)
         {
            final String msg =
               "Only one of 'day-of-week' and 'day-of-month' can be specified for job %s.";
            throw new JobValidationException(String.format(msg, jobName));
         }

         if (dayOfMonth != null && month != null && dayOfMonth != 32)
         {
            Calendar c = Calendar.getInstance();
            c.set(Calendar.MONTH, month);
            int maxDay = c.getActualMaximum(Calendar.DAY_OF_MONTH);

            if (dayOfMonth > maxDay)
            {
               final String msg =
                  "The %d day in month %d must be from the [1..%d] interval for job %s!";
               throw new JobValidationException(String.format(msg, dayOfMonth, month,
                                                              maxDay, jobName));
            }
         }
         
         // month is optional
         validValue(jobName, month, false, "month", 0, 11);
      }
      
      Job job = new Job(jobName, jobType, target, enabled, jobMode);
      job.setMinute(minute);
      job.setHour(hour);
      job.setDayOfWeek(dayOfWeek);
      job.setDayOfMonth(dayOfMonth);
      job.setMonth(month);
      
      return job;
   }
   
   /**
    * Must be called only on server startup.  It starts new {@link #processor job processor} 
    * thread using the server's context. Once the server's sockets have been connected, it reads 
    * the scheduler configuration and schedules the jobs.
    */
   static void scheduleJobs()
   {
      Thread t = new AssociatedThread(processor);
      t.setDaemon(true);
      t.setName(LogHelper.generate("Job processor"));
      t.start();
   }
   
   /**
    * Check if the given value is specified; if specified, check if is in the given [v1 .. v2]
    * interval, inclusive.
    * 
    * @param    jobName
    *           The job name.
    * @param    t
    *           The value to check.  May be <code>null</code>.
    * @param    mandatory
    *           Flag indicating if the value is mandatory.  It will throw a 
    *           {@link JobValidationException} if {@code t} is {@code null} and {@code mandatory}
    *           is {@code true}.
    * @param    descr
    *           A description for this value, to be used in the error message.
    * @param    v1
    *           The left side of the interval.
    * @param    v2
    *           The right side of the interval.
    *           
    * @throws   JobValidationException
    *           If a mandatory value is not specified or if is outside of the given interval.
    */
   private static void validValue(String  jobName, 
                                  Integer t,
                                  boolean mandatory,
                                  String  descr,
                                  int     v1,
                                  int     v2)
   throws JobValidationException
   {
      String msg = null;
      if (mandatory && t == null)
      {
         msg = "'%s' must be specified for job %s.";
      }
      else if (t < v1 || t > v2)
      {
         msg = "%s for job %s is outside interval: [" + v1 + ", " + v2 + "]";
      }

      if (msg != null)
      {
         throw new JobValidationException(String.format(msg, descr, jobName));
      }
   }
   
   /**
    * Read the scheduler configuration and schedule the jobs.
    */
   private synchronized static void initialize()
   {
      if (jobs != null)
      {
         return;
      }
      
      jobs = new HashMap<>();
      
      // access the directory service
      DirectoryService ds = DirectoryService.getInstance();
      
      if (!ds.bind())
         throw new RuntimeException("directory bind failed");

      String schedulerPath = Utils.findDirectoryNodePath(ds, "scheduler", "container", false);
      if (schedulerPath == null)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.logp(Level.WARNING, "Scheduler", "initialize", 
                     LogHelper.generate("No scheduler configured."));
         }

         return;
      }
      
      // enumerate jobs
      String[] jobNodes = ds.enumerateNodes(schedulerPath);
      for (String jobName : jobNodes)
      {
         String jobPath = schedulerPath + "/" + jobName;

         String type   = ds.getNodeString(jobPath, "type");
         String target = ds.getNodeString(jobPath, "target");

         Boolean oenabled   = ds.getNodeBoolean(jobPath, "enabled");
         String  mode       = ds.getNodeString(jobPath,  "mode");
         Integer minute     = ds.getNodeInteger(jobPath, "minute");
         Integer hour       = ds.getNodeInteger(jobPath, "hour");
         Integer dayOfWeek  = ds.getNodeInteger(jobPath, "day-of-Week");
         Integer dayOfMonth = ds.getNodeInteger(jobPath, "day-of-Month");
         Integer month      = ds.getNodeInteger(jobPath, "month");

         Job job = createJob(jobName, type, target, (oenabled != null && oenabled.booleanValue()), 
                             mode, minute, hour, dayOfWeek, dayOfMonth, month);
         
         if (!addJob(job))
         {
            throw new RuntimeException(String.format("Job '%s' is already defined!", jobName));
         }
      }
      
      // last use of the directory, end our session with it
      ds.unbind();
   }
   
   /**
    * Get the next job execution time.
    * 
    * @param    job
    *           The job name.
    * @param    lastExecutionTime
    *           The last time the job was executed, in millis.  If {@code null}, it means it is
    *           the first execution and the current time will be used.
    *           
    * @return   The next execution time, in millis.
    */
   private static long getNextScheduleTime(Job job, Long lastExecutionTime)
   {
      if (lastExecutionTime != null && job.getMode() != JobMode.RECURRENT)
      {
         // only recurrent mode allows more than one time
         return -1;
      }
      
      if (job.getMode() == JobMode.NOW)
      {
         // return a past time, so it will be executed immediately
         return System.currentTimeMillis() - 1000;
      }
      
      if (lastExecutionTime == null)
      {
         // use current time as offset, when scheduling for the first time
         lastExecutionTime = System.currentTimeMillis() + 60 * 1000;
      }
      
      // compute the next execution time
      Calendar c = Calendar.getInstance();
      c.setTimeInMillis(lastExecutionTime - 3600);
      c.setLenient(false);

      c.set(Calendar.MILLISECOND, 0);
      c.set(Calendar.SECOND, 0);
      
      while (true)
      {
         int cminute = c.get(Calendar.MINUTE);
         if (cminute != job.getMinute())
         {
            // (re)adjust the minute
            if (cminute > job.getMinute())
            {
               // advance to next hour if the minute has passed
               int chour = c.get(Calendar.HOUR_OF_DAY) + 1;
               c.set(Calendar.HOUR_OF_DAY, chour);
            }

            c.set(Calendar.MINUTE, job.getMinute());
            continue;
         }
         
         int chour = c.get(Calendar.HOUR_OF_DAY);
         if (job.getHour() != null && job.getHour() != chour)
         {
            // (re)adjust the hour of day

            if (chour > job.getHour())
            {
               c.set(Calendar.HOUR_OF_DAY, job.getHour());
               c.add(Calendar.DAY_OF_MONTH, 1);
               c.set(Calendar.MINUTE, 0);
               continue;
            }
            
            c.set(Calendar.HOUR_OF_DAY, job.getHour());
            c.set(Calendar.MINUTE, 0);
            continue;
         }
         
         int cdayOfWeek = c.get(Calendar.DAY_OF_WEEK);
         if (job.getDayOfWeek() != null && job.getDayOfWeek() != cdayOfWeek)
         {
            // (re)adjust the day of month
            int deltaDays = job.getDayOfWeek() - cdayOfWeek;
            deltaDays += (deltaDays < 0 ? 7 : 0);
            
            c.add(Calendar.DAY_OF_MONTH, deltaDays);
            c.set(Calendar.HOUR_OF_DAY, 0);
            c.set(Calendar.MINUTE, 0);
         }
         
         int cdayOfMonth = c.get(Calendar.DAY_OF_MONTH);
         if (job.getDayOfMonth() != null && job.getDayOfMonth() != cdayOfMonth)
         {
            // (re)adjust the day of day of month
            int jdayOfMonth = (job.getDayOfMonth() == 32) ? job.getDayOfMonth()
                                                          : c.getMaximum(Calendar.DAY_OF_MONTH);
            
            if (cdayOfMonth > jdayOfMonth)
            {
               c.add(Calendar.MONTH, 1);
               c.set(Calendar.DAY_OF_MONTH, 1);
               c.set(Calendar.HOUR_OF_DAY, 0);
               c.set(Calendar.MINUTE, 0);
               continue;
            }
             
            c.set(Calendar.DAY_OF_MONTH, jdayOfMonth);
            c.set(Calendar.HOUR_OF_DAY, 0);
            c.set(Calendar.MINUTE, 0);
            continue;
         }
         
         int cmonth = c.get(Calendar.MONTH);
         if (job.getMonth() != null && cmonth != job.getMonth())
         {
            // (re)adjust the day of month

            if (job.getMonth() > cmonth)
            {
              c.add(Calendar.MONTH, 12 - job.getMonth() + cmonth);
              
              c.set(Calendar.DAY_OF_MONTH, 1);
              c.set(Calendar.HOUR_OF_DAY, 0);
              c.set(Calendar.MINUTE, 0);
              continue;
            }

            c.set(Calendar.MONTH, job.getMonth());
            c.set(Calendar.DAY_OF_MONTH, 1);
            c.set(Calendar.HOUR_OF_DAY, 0);
            c.set(Calendar.MINUTE, 0);
            continue;
         }

         // all found
         break;
      }
 
      return c.getTimeInMillis();
   }
   
   /**
    * Get a runner to execute this job, depending on the job's {@link Job#getType() type}.
    * 
    * @param    job
    *           The job to be executed.
    *           
    * @return   A runnable instance for this job.
    */
   private static Runnable getRunner(Job job)
   {
      switch (job.getType())
      {
         case PROCESS:
            return new ProcessLaunchTask(job.getTarget());
            
         case CLASS:
            return new JavaLaunchTask(job.getTarget(), "run");

         case METHOD:
            String target = job.getTarget();
            int    idx    = target.lastIndexOf(".");
            String cls    = target.substring(0, idx);
            String method = target.substring(idx + 1);
            
            return new JavaLaunchTask(cls, method);
      }
      
      return null;
   }
   
   /**
    * Launcher for java-style tasks.
    */
   private static class JavaLaunchTask
   implements Runnable
   {
      /** The target class. */
      private final String cls;
      
      /** The target method. */
      private final String method;
      
      /**
       * Configure this launcher.
       * 
       * @param    cls
       *           The target class.
       * @param    method
       *           The target method.
       */
      public JavaLaunchTask(String cls, String method)
      {
         this.cls = cls;
         this.method = method;
      }

      /**
       * Invoke the {@link #method} from the specified {@link #cls class}.  If the method is not 
       * static, a new instance of {@link #cls} type will be created.
       */
      @Override
      public void run()
      {
         try
         {
            Class<?> clazz = Class.forName(cls);
            Method mthd = clazz.getMethod(method);
            Object instance = (Modifier.isStatic(mthd.getModifiers())) ? null 
                                                                       : clazz.newInstance();
            
            mthd.invoke(instance);
         }
         catch (Exception e)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               final String msg = String.format("Could not launch java job %s.%s", cls, method);
               LOG.logp(Level.SEVERE, "JavaLaunchTask", "run", msg);
            }
         }
      }
   }
   
   /**
    * Implementation for launching a P2J process; this class must be executed under the server's
    * context.  If the P2J process is linked with an appserver, it will delegate the launching to
    * {@link AppServerLauncher#launch}.  Else, it will use the {@link ProcessClientSpawner 
    * spawner} infrastructure to start a P2J client.
    */
   private static class ProcessLaunchTask
   implements Runnable
   {
      /** The P2J process subject ID. */
      private final String process;

      /**
       * Create a new launcher task for the given process.
       * 
       * @param    process
       *           The process subject ID.
       */
      public ProcessLaunchTask(String process)
      {
         this.process = process;
      }
      
      /**
       * Calls {@link ProcessClientSpawner#launch(String)} to launch this P2J {@link #process}.
       */
      @Override
      public void run()
      {
         ProcessClientSpawner.launch(process);
      }
      
      /**
       * Get the string representation of this runner.
       */
      @Override
      public String toString()
      {
         return String.format("Launcher for process '%s'.", process);
      }
   }
   
   /**
    * Implements tasks to be passed to the java {@link #timer}.
    * <p>
    * When a job is executed, it will delegate the job execution to the 
    * {@link Scheduler#processor job processor}, which will execute the job using the server's
    * context. 
    */
   private static class JobTask
   extends TimerTask
   {
      /** The job name. */
      private final String jobName;
      
      /**
       * Create a new job task to be registered with the {@link #timer}.
       */
      public JobTask(String jobName)
      {
         this.jobName = jobName.toLowerCase();
      }

      /**
       * Execute a the job with the given {@link #jobName name}.
       */
      @Override
      public void run()
      {
         // we are executing, remove it from the map
         synchronized (jobRegistry)
         {
            jobRegistry.remove(jobName);
         }

         // get the configured job, reschedule it (if needed) and execute it
         synchronized (jobs)
         {
            Job job = getJob(jobName);
            
            if (!job.isEnabled())
            {
               // the job is not enabled.
               if (LOG.isLoggable(Level.WARNING))
               {
                  String msg = String.format("Can not reschedule: job %s is disabled.", jobName);
                  LOG.logp(Level.WARNING, "Scheduler$JobTask", "run", LogHelper.generate(msg));
                  return;
               }
            }

            // schedule it again (depending on the mode)
            scheduleJob(jobName, this.scheduledExecutionTime());
            
            if (LOG.isLoggable(Level.FINE))
            {
               String msg = String.format("Executing job %s at %tc.",
                                          jobName, this.scheduledExecutionTime());
               LOG.logp(Level.FINE, "Scheduler$JobTask", "run", LogHelper.generate(msg));
            }
            
            // get the core executor
            Runnable core = getRunner(job);
            
            processor.executeJob(core);
         }
      }
   }
   
   /**
    * Specialized class which listens for incoming jobs (sent by the {@link #timer}.  
    * This class runs in a separate thread, under the server's context.
    */
   private static class JobProcessor
   implements Runnable
   {
      /** The list with the pending jobs. */
      private final LinkedList<Runnable> pendingJobs = new LinkedList<>();
      
      /**
       * Initialize the scheduler and schedule the initial jobs.  Then, start listening for new
       * jobs being posted to the {@link #pendingJobs queue} by the {@link #timer}.
       */
      @Override
      public void run()
      {
         initialize();
         
         // before starting, wait for the server to start listening
         try
         {
            SessionManager.get().waitUntilReady();
         }
         catch (InterruptedException e)
         {
            return;
         }
         
         // add initial jobs
         synchronized (jobs)
         {
            for (Job job : jobs.values())
            {
               if (!job.isEnabled())
               {
                  // skip disabled jobs
                  continue;
               }
               
               scheduleJob(job.getName(), null);
            }
         }
         
         Runnable core = null;

         // listen for incoming jobs
         while (true)
         {
            synchronized (pendingJobs)
            {
               try
               {
                  pendingJobs.wait();
               }
               catch (InterruptedException e)
               {
                  // someone interrupted me, exit
                  break;
               }
               
               core = (pendingJobs.isEmpty()) ? null : pendingJobs.removeFirst();
            }
            
            if (core != null)
            {
               // execute this in a different thread, but with the server's context
               Thread t = new AssociatedThread(core);
               String tname = LogHelper.createThreadName(String.format(core.toString()));
               t.setDaemon(true);
               t.setName(tname);
               t.start();
            }
         }
      }
      
      /**
       * Execute the given job.
       * 
       * @param    core
       *           The {@link Runnable} instance to execute the job.
       */
      public void executeJob(Runnable core)
      {
         synchronized (pendingJobs)
         {
            pendingJobs.addLast(core);
            pendingJobs.notify();
         }
      }
   }
}
