/*
** Module   : ServerKeyStore.java
** Abstract : define a container used to export server key store.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 MAG 20131120 First version.
** 002 CA  20140207 If not explicitly specified, build the server's private key from the directory.
*/

package com.goldencode.p2j.main;

import java.io.*;
import java.util.logging.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.util.*;

/**
 * A container used to export server key store.
 * The server KeyStore is loaded into this structure and serialized (exported)
 * via a RemoteObject protocol to the spawned process.
 * 
 * @author mag
 *
 */
public class ServerKeyStore
implements Serializable
{
   /** Serial Version ID */
   private static final long serialVersionUID = -3745063190770323563L;

   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ServerKeyStore.class.getName());
   
   /** Key store file content */
   private byte[] keyStore;

   /** Key store password */
   private String keyStorePassword;

   /** Key store manager password */
   private String keyManagerPassword;

   /**
    * Constructor. Export server KeyStore.
    * 
    * @param   config
    *          Server bootstrap configuration.
    */
   public ServerKeyStore(BootstrapConfig config) 
   {
      // key store file name
      String keyStoreFile = config.getString("security", "keystore", "filename", null);
      
      if (keyStoreFile != null)
      {
         // read key store content
         File file = new File(keyStoreFile);
         keyStore = new byte[(int) file.length()];
         DataInputStream dis;
         try
         {
            dis = new DataInputStream(new FileInputStream(file));
            dis.readFully(keyStore);
            dis.close();
            // get passwords
            keyStorePassword = config.getString("access", "password", "keystore", null);
            keyManagerPassword = config.getString("access", "password", "keyentry", null);
         }
         catch (IOException e)
         {
            LOG.logp(Level.SEVERE, "ServerKeyStore", "",
                     LogHelper.generate("Cannot load server keystore file."), e);
         }
      }
      else
      {
         // build the key store from the directory using the server's alias
         SecurityManager sm = SecurityManager.getInstance();
         String alias = sm.getServerAlias();
         
         String ksPassword = RandomWordGenerator.create();
         String kePassword = RandomWordGenerator.create();
         this.keyStore = sm.getEncryptedKeyStore(alias, ksPassword, kePassword);
         if (this.keyStore == null)
         {
            this.keyManagerPassword = null;
            this.keyStorePassword = null;
            this.keyStore = null;
         }
         else
         {
            this.keyStorePassword = ksPassword;
            this.keyManagerPassword = kePassword;
         }
      }
   }

   /**
    * Get KeyStore file content.
    * 
    * @return  KeyStore file content.
    */
   public byte[] getKeyStore()
   {
      return keyStore;
   }

   /**
    * Get KeyStore password.
    * 
    * @return  KeyStore password.
    */
   public String getKeyStorePassword()
   {
      return keyStorePassword;
   }
   
   /**
    * Get KeyStore manager password.
    * 
    * @return  KeyStore manager password.
    */
   public String getKeyManagerPassword()
   {
      return keyManagerPassword;
   }
   
   /**
    * Check if the content is valid
    * 
    * @return  <code>true</code> if all parameters are not <code>null</code>
    *          <code>false</code> otherwise.
    */
   public boolean isValid()
   {
      return keyStore != null && keyStorePassword != null && keyManagerPassword != null;
   }
}
