/*
** Module   : SpawnerImpl.java
** Abstract : spawn the web client.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 MAG 20131118 First version.
** 002 CA  20140206 Refactored to make it unaware of the type of the P2J client beeing spawned.
*/

package com.goldencode.p2j.main;

import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.util.*;

/**
 * This class implements the Spawner contract exported via a RemoteObject protocol.
 * This API is used by the P2J client spawned process to communicate with the P2J server side.
 * 
 * @author mag
 * 
 */
public class SpawnerImpl
implements Spawner
{
   /** Logger */
   @SuppressWarnings("unused")
   private static final Logger LOG = LogHelper.getLogger(SpawnerImpl.class.getName());

   /** Instance of Spawner */
   private static final Spawner spawner = new SpawnerImpl();
   
   /** Registered server listeners */
   private Map<String, SpawnerListener> listeners = new ConcurrentHashMap<>();
   
   /**
    * Get the Spawner instance.
    * 
    * @return  The instance of spawner.
    */
   public static Spawner getInstance()
   {
      return spawner;
   }
   
   /**
    * Register a Remote Object Server.
    */
   private SpawnerImpl()
   {      
      RemoteObject.registerNetworkServer(Spawner.class, this);
   }
      
   /**
    * Notify when the client has started.
    * 
    * @param    uuid
    *           Remote client assigned identifier.
    * @param    data
    *           Custom data sent by the P2J client back to the server.
    */
   @Override
   public void clientIsReady(String uuid, Object data)
   {
      SpawnerListener listener = listeners.get(uuid);
      if (listener != null)
      {
         listener.clientIsReady(data);
      }
   }

   /**
    * Depending on the implementation, export custom data from the server back to the client.
    * 
    * @param    uuid
    *           Remote client assigned identifier.
    *
    * @return   Custom data sent by the server back to the client.
    */
   @Override
   public Object exportData(String uuid)
   {
      SpawnerListener listener = listeners.get(uuid);
      return listener == null ? null : listener.getServerData();
   }

   /**
    * Get a {@link TemporaryClient} worker which will do the work after authenticating on the P2J
    * server using the temporary credentials.
    * 
    * @param    uuid
    *           Remote client assigned identifier.
    * 
    * @return   See above.
    */
   @Override
   public TemporaryClient getTemporaryClient(String uuid)
   {
      return listeners.get(uuid).getTemporaryClient();
   }

   /**
    * Add listener.
    * 
    * @param   uuid
    *          Remote client assigned identifier.
    * @param   listener
    *          Listener instance to add.
    */
   public void addListener(String uuid, SpawnerListener listener)
   {
      listeners.put(uuid, listener);
   }

   /**
    * Remove listener.
    * 
    * @param   uuid
    *          Remote client assigned identifier.
    */
   public void removeListener(String uuid)
   {
      listeners.remove(uuid);
   }
}
