/*
** Module   : ProcessAccount.java
** Abstract : represents an application or server process account
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description--------------------------------
** 001 NVS 20050307   @20199 Created. Process account extends account and
**                           represents P2j application accounts. 
** 002 NVS 20050311   @20285 Constructor now accepts Booleans which can be
**                           nulls.
** 003 NVS 20050311   @20296 Methods scope changed to package private.
** 004 GES 20081104   @40340 Moved common code into the base class.
** 005 NVS 20090603   @42591 Provide enabled account status info for super().
** 006 CA  20130529          Added appserver support.
** 007 CA  20140206          Changed H006: appserver agents can be started only from P2J processes.
*/ 

package com.goldencode.p2j.security;

/**
 * Represents an application or server process account.  This is a
 * non-interactive account, but it can login and it can be assigned specific
 * rights as a valid entity in the system.
 * <p>
 * WARNING: at this time process accounts cannot be assigned to groups.  This
 * is something that will be fixed in a future enhancement. 
 */
class ProcessAccount
extends Account
{
   /** Server or application flag. */
   private boolean isServer = false;

   /** Master Server flag. */
   private boolean isMaster = false;
   
   /** The name of the app server. May be <code>null</code>. */
   private String appServer = null;

   /**
    * Constructor.
    *
    * @param    subjectId
    *           subject ID also known as account name
    * @param    enabled
    *           <code>true</code> if this account is enabled
    * @param    description
    *           text that describes the purpose of this application; may be <code>null</code>
    * @param    alias
    *           truststore alias associated with this application's certificate; 
    *           may be <code>null</code>
    * @param    isServer
    *           flag indicating server vs application; may be <code>null</code>
    * @param    isMaster
    *           flag indicating master server; may be <code>null</code>
    * @param    appServer
    *           The name of the App Server. May be <code>null</code>.
    */
   ProcessAccount(String  subjectId,
                  boolean enabled,
                  String  description,
                  String  alias, 
                  Boolean isServer,
                  Boolean isMaster,
                  String  appServer)
   {
      super(subjectId, enabled, description, alias);
      
      this.appServer = appServer;

      if (isServer == null)
      {
         this.isServer = false;
         this.isMaster = false;
      }
      else
      {
         this.isServer = isServer.booleanValue();
         if (this.isServer)
            this.isMaster = isMaster.booleanValue();
         else
            this.isMaster = false;
      }
   }

   /**
    * Gets the account type of this account
    *
    * @return   The <code>ACC_PROCESS</code> constant.
    */
   int getAccountType()
   {
      return ACC_PROCESS;
   }

   /**
    * Gets server or application flag
    *
    * @return <code>true</code> if it is a server
    */
   boolean isServer()
   {
      return isServer;
   }

   /**
    * Gets master server flag
    *
    * @return <code>true</code> if it is a master server
    */
   boolean isMaster()
   {
      return isMaster;
   }
   
   /**
    * Get the appserver name associated with this account.
    * 
    * @return    The {@link #appServer} name.
    */
   String getAppServer()
   {
      return appServer;
   }
}
