/*
** Module   : SSLCertGenUtil.java
** Abstract : Utility to generate the root CA and SSL certificates for all the accounts in a
**            directory.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140207 First version.
*/

package com.goldencode.p2j.security;

import java.io.*;
import java.math.*;
import java.security.*;
import java.security.cert.Certificate;
import java.security.cert.*;
import java.util.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.util.*;

/**
 * Generate new root CA and SSL certificates for all the accounts which authenticate via a 
 * certificate and save them and the associated private keys in the directory.  All existing CAs,
 * peer certificates and private keys will be removed.
 * <p>
 * The private keys will be saved under the "/security/certificates/private-keys/" node.
 * <p>
 * The company information (used to build the certificate's subject) will be saved under the
 * "/security/certificates/company/" node.  If the node does not already exist, it will ask the
 * user for the information.
 */
public class SSLCertGenUtil
{
   /** Constant identifying the CAs node in the directory. */
   private static final String ROOT_CA_NODE = "/security/certificates/cas/";

   /** Constant identifying the peer certs node in the directory. */
   private static final String ROOT_PEER_NODE = "/security/certificates/peers/";
   
   /** Constant identifying the security/certificates node in the directory. */
   private static final String ROOT_CERTIFICATES_NODE = "/security/certificates";

   /** Constant identifying the process accounts node in the directory. */
   private static final String ROOT_PROCESSES_NODE = "/security/accounts/processes/";

   /** Constant identifying the user accounts node in the directory. */
   private static final String ROOT_USERS_NODE = "/security/accounts/users/";
   
   /** Constant identifying the company information node in the directory. */
   private static final String ROOT_COMPANY_NODE = "/security/certificates/company/";
   
   /** Constant identifying the private keys container node in the directory. */
   private static final String ROOT_PRIVATE_KEYS_NODE = "/security/certificates/private-keys/";

   /** Constant identifying the name of the company node containing the certificate validity. */
   private static final String CERT_VALIDITY_YEARS_NODE = "validityYears";

   /**
    * Constant identifying the name of the company node containing the RSA private key strength.
    */
   private static final String RSA_PRIVATE_KEY_STRENGTH = "rsaKeyStrength";

   /** Constant identifying the name of the company node containing the RSA public exponent. */
   private static final String RSA_PUBLIC_EXPONENT = "rsaPublicExponent";
   
   /** Read data from the standard input. */
   private final BufferedReader reader;

   /** Configuration used to initialize the directory. */
   private final BootstrapConfig cfg;
   
   /** Map identifying the company attributes. */
   private final Map<String, String> company = new HashMap<>();
   
   /** All the collected aliases from the defined accounts. */
   private final Set<String> aliases = new TreeSet<>();
   
   /** All the aliases associated with server processes. */
   private final Set<String> serverAliases = new TreeSet<>();
   
   /** Factory to generate the SSL certificate. */
   private final SSLCertFactory factory;
   
   /** Store where to add the trusted certificates (servers and CAs). */
   private final KeyStore trustCertStore;
   
   /** Store where to add the private keys for the trusted certificates. */
   private final KeyStore trustKeyStore;

   /** Store where to add the certificates associated with non-server accounts. */
   private final KeyStore accCertStore;

   /**
    * Store where to add the private keys for the certificates associated with non-server 
    * accounts.
    */
   private final KeyStore accKeyStore;

   /**
    * The random passwords used to encrypt the private keys in their {@link KeyStore}, per each
    * alias.
    */
   private final Map<String, String> keyEntryPasswords = new HashMap<>();
   
   /** The certificate's validity, in years. */
   private int validity = -1;

   /** Alias for the root CA. */
   private String rootCAAlias = null;
   
   /** Instance to access the directory. */
   private DirectoryService ds;
   
   /** The master password to encrypt all private keys, in the directory. */
   private String masterPassword = null;

   /**
    * Create a new utility.
    * 
    * @param    config
    *           Configuration for accessing the directory.
    *           
    * @throws   SSLCertGenException
    *           If the utility could not be instantiated.
    */
   public SSLCertGenUtil(BootstrapConfig config)
   throws SSLCertGenException
   {
      Console console = System.console();
      
      if (console != null)
      {
         reader = new BufferedReader(console.reader());
      }
      else
      {
         reader = new BufferedReader(new InputStreamReader(System.in));
      }
      
      cfg = new BootstrapConfig(config);
      if (cfg.getString("directory", "xml", "filename", null) == null)
      {
         throw new NullPointerException("The directory file must be specified!");
      }
      
      // hard-coded to Bouncy Castle
      this.factory = new BCCertFactory();

      // we are using 4 KeyStore instances:
      // 1. key store for the private key for the self-signed, root CA certificate
      // 2. trust-store with the trusted certificates (root CA certificate + server certificates)
      // 3. key store for the private key for each non-serverprocess and user accounts
      // 4. cert store with the process and user certificates

      trustCertStore = createEmptyStore();
      trustKeyStore  = createEmptyStore();
      accCertStore  = createEmptyStore();
      accKeyStore   = createEmptyStore();
   }
   
   /**
    * Main method to (re)generate the root CA, peer certificates and private keys.
    * 
    * @throws    ConfigurationException
    *            In case the {@link #ds directory service} could not be initialized.
    * @throws    IOException
    *            If standard input can not be accessed.
    * @throws    SSLCertGenException
    *            In case of problems during the generation of root CA or peer certificates.
    */
   public void generate()
   throws ConfigurationException,
          IOException,
          SSLCertGenException
   {
      cfg.setConfigItem("directory", "xml", "must_exist", "true");

      String dirFile = cfg.getString("directory", "xml", "filename", null);
      System.out.println("Initializing service for directory " + dirFile + "...");
      this.ds = DirectoryService.createInstance(cfg);

      if (!ds.bind())
         throw new RuntimeException("bind() failed");

      // step 1. read configuration from the /security/certificates/company node
      //         any missing nodes are read from user input and saved
      System.out.println("Reading company configuration...");
      readCompanyConfiguration();

      // step 2. read processes (servers and batch) and users configured with a alias
      readAccounts(ROOT_PROCESSES_NODE);
      readAccounts(ROOT_USERS_NODE);
      
      // step 3. generate root CA and the certificates for each account
      generateRootCA();
      generateCertificates();
      
      // check how to encrypt the private keys
      String answer = null;
      do
      {
         answer = readLine("Encrypt the private keys using the same password (yes/no): ");
      }
      while (!("no".equalsIgnoreCase(answer) || "yes".equalsIgnoreCase(answer)));
      
      masterPassword = ("yes".equals(answer)) ? createAES256BitKey() : null;
      
      // step 4. save the certificates and the keys in the directory
      saveRootCA();
      saveCertificates();
      
      if (!ds.unbind())
         throw new RuntimeException("unbind() failed");
 
      // save server's certificates in an external key store
      answer = null;
      do
      {
         answer = readLine("Save servers' certificates in an external key store (yes/no): ");
      }
      while (!("no".equalsIgnoreCase(answer) || "yes".equalsIgnoreCase(answer)));
      
      if ("yes".equalsIgnoreCase(answer))
      {
         String file = null;
         do
         {
            file = readLine("Enter key store file name: ");
         }
         while (file.length() == 0);
         
         try
         {
            KeyStore store = createEmptyStore();
            for (String alias : serverAliases)
            {
               Certificate cert = trustCertStore.getCertificate(alias);
               
               store.setCertificateEntry(alias, cert);
            }
            String pass = RandomWordGenerator.create();
            store.store(new FileOutputStream(new File(file)), pass.toCharArray());
            
            String msg = "The servers' certificates %s were saved in the %s key store, " +
                         "using password '%s'.";
            System.out.println(String.format(msg, serverAliases.toString(), file, pass));
         }
         catch (KeyStoreException | NoSuchAlgorithmException | CertificateException e)
         {
            throw new SSLCertGenException(e);
         }
      }
      
      System.out.println("Done.");
      System.out.println("WARNING! Any configuration set at the client.xml or server.xml files " +
                         "or via bootstrap config arguments will have priority over the " +
                         "in-directory keys or certificates.");
      System.out.println("WARNING! The private key encryption passwords for all the account " +
                         "certificates are saved unencrypted in the directory.  The root CA's " +
                         "private key can be safely deleted, if it is not required to issue " +
                         "other certificates using this CA.");
      if (masterPassword != null)
      {
         System.out.println("WARNING! All private keys are encrypted using the same [" +
                            masterPassword + "] password. If needed, delete the 'key-password' " +
                            "nodes from the directory manually, and set this password using the " +
                            "access:password:masterkeyentry bootstrap config at server startup.");
      }
   }
   
   /**
    * Save all the root CA certificate and its private key in the directory.
    * 
    * @throws    SSLCertGenException
    *            If the data could not be saved.
    */
   private void saveRootCA()
   throws SSLCertGenException
   {
      openBatch(ROOT_CERTIFICATES_NODE);
      
      savePrivateKey(rootCAAlias, trustKeyStore);
      saveCertificate(rootCAAlias, trustCertStore, ROOT_CA_NODE);
      
      closeBatch(ROOT_CERTIFICATES_NODE);
   }
   
   /**
    * Save all the peer certificates and their private keys in the directory.
    * 
    * @throws    SSLCertGenException
    *            If the data could not be saved.
    */
   private void saveCertificates()
   throws SSLCertGenException
   {
      openBatch(ROOT_CERTIFICATES_NODE);

      for (String alias : aliases)
      {
         boolean isServer = serverAliases.contains(alias);
         
         savePrivateKey(alias, (isServer ? trustKeyStore : accKeyStore));
      }

      for (String alias : aliases)
      {
         boolean isServer = serverAliases.contains(alias);
         
         saveCertificate(alias, (isServer ? trustCertStore : accCertStore), ROOT_PEER_NODE);
      }

      closeBatch(ROOT_CERTIFICATES_NODE);
   }
   
   /**
    * Save the certificate for the specified alias in the directory, under the specified parent
    * node.
    * 
    * @param    alias
    *           The alias.
    * @param    certStore
    *           The store from which to retrieve the certificate.
    * @param    parentNode
    *           The parent node where to save the certificate.
    *           
    * @throws   SSLCertGenException
    *           If the data could not be saved.
    */
   private void saveCertificate(String alias, KeyStore certStore, String parentNode)
   throws SSLCertGenException
   {
      String fullId = parentNode + alias;
      deleteNode(fullId);
      
      System.out.println(String.format("Adding certificate for alias [%s] to the %s node...",
                                       alias, fullId));
      
      try
      {
         Certificate cert = certStore.getCertificate(alias);

         addNode(fullId, "bytes", cert.getEncoded());
      }
      catch (KeyStoreException | CertificateEncodingException e)
      {
         throw new SSLCertGenException(e);
      }
      
   }
   
   /**
    * Save the private key for the specified alias in the directory.
    * <p>
    * The private key is encrypted using a random password, which will be saved too in the
    * directory.
    * 
    * @param    alias
    *           The alias.
    * @param    privateKeyStore
    *           The store from which to read the private key.
    *           
    * @throws   SSLCertGenException
    *           If the private key could not be saved.
    */
   private void savePrivateKey(String alias, KeyStore privateKeyStore)
   throws SSLCertGenException
   {
      String fullId = ROOT_PRIVATE_KEYS_NODE + alias;

      if (ds.getNodeClass(ROOT_PRIVATE_KEYS_NODE) == null)
      {
         addNode(ROOT_PRIVATE_KEYS_NODE, "container", null);
      }
      
      deleteNode(fullId);
      addNode(fullId, "container", null);
      
      String password = keyEntryPasswords.get(alias);
      boolean rootCA = rootCAAlias.equals(alias);
      
      // determine which password to use: master or distinct
      String dirPassword = (masterPassword != null && !rootCA)
                              ? masterPassword 
                              : createAES256BitKey();
      if (!rootCA)
      {
         // do not save the root CA password
         addNode(ROOT_PRIVATE_KEYS_NODE + alias + "/key-password", "bytes", dirPassword.getBytes());
      }
      
      try
      {
         Key key = privateKeyStore.getKey(alias, password.toCharArray());

         byte[] encrypted = factory.encryptPrivateKey(key, dirPassword);
         addNode(ROOT_PRIVATE_KEYS_NODE + alias + "/key-entry", "bytes", encrypted);
      }
      catch (UnrecoverableKeyException | KeyStoreException | NoSuchAlgorithmException e)
      {
         throw new SSLCertGenException(e);
      }
      
      final String msg1 = "Private key for alias [%s] was encrypted using the [%s] password " +
                          "and saved in the %s node.\nUse the access:password:keyentry-%s key " +
                          "to set this password in the server's bootstrap config file and " +
                          "delete  the /security/certificates/private-keys/%s node from " +
                          "the directory.";
      final String msg2 = "Private key for root CA [%s] was encrypted using the [%s] password " +
                          "and the password was NOT SAVED in the directory.  Losing this " +
                          "password will prohibit using the root CA to issue new certificates.";
      if (rootCA)
      {
         System.out.println(String.format(msg2, alias, dirPassword));
      }
      else
      {
         System.out.println(String.format(msg1, alias, dirPassword, fullId, alias, alias));
      }
   }
   
   /**
    * Generate peer certificates for all {@link #aliases}.
    * 
    * @throws    SSLCertGenException
    *            If the SSL certificates could not be generated.
    */
   private void generateCertificates()
   throws SSLCertGenException
   {
      for (String alias : aliases)
      {
         boolean isServer = (serverAliases.contains(alias));
         String type = (isServer ? "server" : "account");
         String commonName = alias + " " + type;
         
         System.out.println(String.format("Generating certificate for %s with alias [%s]...",
                                          type, alias));
         
         String password = factory.generateCertificate(alias, validity, 
                                                       commonName, company,
                                                       (isServer ? trustCertStore : accCertStore), 
                                                       (isServer ? trustKeyStore : accKeyStore));

         keyEntryPasswords.put(alias, password);
      }
   }
   
   /**
    * Generate the root CA.
    * 
    * @throws    SSLCertGenException
    *            If the root CA could not be generated.
    * @throws    IOException
    *            If the alias for the root CA could not be read from standard input.
    */
   private void generateRootCA()
   throws SSLCertGenException,
          IOException
   {
      // find the root CA alias
      openBatch(ROOT_CA_NODE);
      
      String[] cas = ds.enumerateNodes(ROOT_CA_NODE);
      if (cas.length == 0)
      {
         // read the root CA alias
         while (rootCAAlias == null || rootCAAlias.length() == 0)
         {
            rootCAAlias = readLine("Enter the root CA alias: ");
            
            if (!aliases.contains(rootCAAlias))
            {
               break;
            }
            
            System.out.println("Alias " + rootCAAlias + " is in use!");
            rootCAAlias = null;
         }
      }
      else
      {
         // reuse the root CA alias
         rootCAAlias = cas[0];
      }

      closeBatch(ROOT_CA_NODE);
      
      final String msg = "Generating root CA using [%s] alias...";
      System.out.println(String.format(msg, rootCAAlias));
      
      String password = factory.generateRootCA(rootCAAlias, validity,
                                               rootCAAlias + " - root CA authority", company,
                                               trustCertStore, trustKeyStore);
      
      keyEntryPasswords.put(rootCAAlias, password);
   }
   
   /**
    * Read all the accounts which have an alias specified.
    * 
    * @param    rootNode
    *           The node from which to read the accounts.
    */
   private void readAccounts(String rootNode)
   {
      openBatch(rootNode);
      
      String[] accIds = ds.enumerateNodes(rootNode);
      if (accIds.length == 0)
      {
         System.out.println("No accounts defined in " + rootNode + "!");
         return;
      }
      
      String  alias    = null;
      Boolean isServer = null;
      
      // process existing nodes
      for (int i = 0; i < accIds.length; i++)
      {
         String accID = accIds[i];
         String fullId = new String(rootNode + accID);
         String nodeCls = ds.getNodeClass(fullId);
         
         if ("/meta/class/process".equals(nodeCls))
         {
            isServer = ds.getNodeBoolean(fullId, "server");
         }
         else if ("/meta/class/user".equals(nodeCls))
         {
            // user account
         }
         else
         {
            System.out.println("Account " + fullId + " ignored - not a user or process");
            continue;
         }

         alias    = ds.getNodeString(fullId,  "alias");
         
         if (alias == null || alias.length() == 0)
         {
            System.out.println("Account " + fullId + " ignored - no alias is defined");
            continue;
         }

         if (aliases.contains(alias))
         {
            System.out.println("WARNING: alias " + alias + " is set for multiple accounts!");
            continue;
         }

         System.out.println("Adding account " + fullId);

         aliases.add(alias);
         
         if (isServer != null && isServer.booleanValue())
         {
            serverAliases.add(alias);
         }
      }
      
      closeBatch(rootNode);
   }
   
   /**
    * Read the existing company configuration from the directory or, if missing, read it from
    * the standard input.
    * 
    * @throws    IOException
    *            In case of problems during reading.
    * @throws    SSLCertGenException
    *            If the {@link SSLCertFactory} factory could not be initialized.
    */
   private void readCompanyConfiguration()
   throws IOException,
          SSLCertGenException
   {
      Map<String, String> mandatoryKeys = factory.getMandatorySubjectFields();
      
      while (!ds.openBatch(ROOT_COMPANY_NODE))
      {
         // node does not exist, add it
         openBatch(ROOT_CERTIFICATES_NODE);
         
         // add a new node
         addNode(ROOT_COMPANY_NODE, "container", null);
         
         closeBatch(ROOT_CERTIFICATES_NODE);
      }

      Integer rsaKeyStrength = null;
      BigInteger rsaPublicExponent = null;
      
      // read the existing nodes first
      String[] existing = ds.enumerateNodes(ROOT_COMPANY_NODE);
      for (int i = 0; i < existing.length; i++)
      {
         String oid = existing[i].toUpperCase();
         
         if (oid.equalsIgnoreCase(CERT_VALIDITY_YEARS_NODE))
         {
            validity = ds.getNodeInteger(ROOT_COMPANY_NODE + oid, "value");
            System.out.println(String.format("Using %d years for validity.", validity));
            continue;
         }
         
         if (oid.equalsIgnoreCase(RSA_PRIVATE_KEY_STRENGTH))
         {
            rsaKeyStrength = ds.getNodeInteger(ROOT_COMPANY_NODE + oid, "value");
            System.out.println(String.format("Using %d bytes for RSA private key strength.", 
                                             rsaKeyStrength));
            continue;
         }
         
         if (oid.equalsIgnoreCase(RSA_PUBLIC_EXPONENT))
         {
            // the exponent can be a very large or even a hex value, so keep it in a string. 
            String exp = ds.getNodeString(ROOT_COMPANY_NODE + oid, "value");
            rsaPublicExponent = new BigInteger(exp);
            System.out.println(String.format("Using %s for RSA public exponent.", exp));
            continue;
         }
         
         String val = ds.getNodeString(ROOT_COMPANY_NODE + oid, "value").trim();

         // existing value (may be empty)
         String txt = mandatoryKeys.get(oid);
         System.out.println(String.format("Using '%s' for [%s] %s.", val, oid,
                                          (txt == null ? "" : txt)));
         
         mandatoryKeys.remove(oid);
         company.put(oid, val);
      }
      
      // get user input for the remainder of the nodes
      for (String oid : mandatoryKeys.keySet())
      {
         String txt = mandatoryKeys.get(oid);
         
         // read the value
         String val = readLine(String.format("%s [%s]: ", txt, oid));
         
         addNode(ROOT_COMPANY_NODE + oid, "string", val);
         
         // save empty values
         System.out.println(String.format("Using '%s' for [%s] %s.", val, oid, txt));
      }

      // check the certificate validity
      if (validity == -1)
      {
         while (true)
         {
            try
            {
               validity = Integer.parseInt(readLine("Enter validity (years): "));
               
               if (validity <= 0)
               {
                  System.out.println("Enter a number of years greater than zero.");
                  continue;
               }
               
               break;
            }
            catch (NumberFormatException e)
            {
               System.out.println("Incorrect year.");
               continue;
            }
         }
         
         addNode(ROOT_COMPANY_NODE + CERT_VALIDITY_YEARS_NODE, "integer", validity);
         
         System.out.println(String.format("Using %d years for validity.", validity));
      }
      
      // get the key strength
      if (rsaKeyStrength == null)
      {
         String line = readLine("Enter RSA key strength (in bytes), blank for default (1024): ");
         rsaKeyStrength = (line.length() > 0) ? new Integer(line) : 1024;
         
         if (rsaKeyStrength > 1024)
         {
            addNode(ROOT_COMPANY_NODE + RSA_PRIVATE_KEY_STRENGTH, "integer", rsaKeyStrength);
            
            System.out.println(String.format("Using %d bytes for RSA private key strength.", 
                                             rsaKeyStrength));
         }
      }
      // get the public exponent
      if (rsaPublicExponent == null)
      {
         String line = readLine("Enter RSA public exponent, blank for default (65337): ");
         if (line.length() > 0)
         {
            rsaPublicExponent = new BigInteger(line);

            addNode(ROOT_COMPANY_NODE + RSA_PUBLIC_EXPONENT,
                    "string", rsaPublicExponent.toString());
            
            System.out.println(String.format("Using %s bytes for RSA private key strength.", 
                                             rsaPublicExponent.toString()));
         }
         else
         {
            rsaPublicExponent = new BigInteger("65337");
         }
      }
      
      factory.init(rsaKeyStrength, rsaPublicExponent);
      
      closeBatch(ROOT_COMPANY_NODE);
   }

   /**
    * Delete the specified node and all of its children.
    * 
    * @param    nodeId
    *           The node ID.
    * 
    * @throws   RuntimeException
    *           If the node or one of its children could not be deleted.
    */
   private void deleteNode(String nodeId)
   {
      if (ds.getNodeClass(nodeId) != null)
      {
         String[] children = ds.enumerateNodes(nodeId);
         for (int i = 0; i < children.length; i++)
         {
            String fullId = nodeId + "/" + children[i];
            
            // remove previous entry
            if (!ds.deleteNode(fullId))
            {
               throw new RuntimeException("deleteNode() failed for the " + fullId + " node");
            }
         }

         if (!ds.deleteNode(nodeId))
         {
            throw new RuntimeException("deleteNode() failed for the " + nodeId + " node");
         }
      }
   }
   
   /**
    * Add a new node to the directory, having the specified class and a single attribute
    * with the same class and specified value.
    * 
    * @param    nodeId
    *           The node ID.
    * @param    cls
    *           The node's and attribute's class.
    * @param    value
    *           The attribute's value.
    * 
    * @throws   RuntimeException
    *           If the node could not be added.
    */
   private void addNode(String nodeId, String cls, Object value)
   {
      NodeAttribute nodeAttr = ds.getClassNodeAttribute(cls, "value");
      Attribute attr = new Attribute(nodeAttr, new Object[] { value });
      addNode(nodeId, cls, new Attribute[] { attr });
   }
   
   /**
    * Add a new node to the directory, having the specified class and attributes.
    * 
    * @param    nodeId
    *           The node ID.
    * @param    cls
    *           The node's class.
    * @param    attrs
    *           The node's attributes.
    * 
    * @throws   RuntimeException
    *           If the node could not be added.
    */
   private void addNode(String nodeId, String cls, Attribute[] attrs)
   {
      if (!ds.addNode(nodeId, cls, attrs))
      {
         throw new RuntimeException("addNode() failed for the " + nodeId + " node");
      }
   }
   
   /**
    * Open a batch editing session for the specified node.
    * 
    * @param    node
    *           The node ID.
    * 
    * @throws   RuntimeException
    *           If the batch editing session could not be opened.
    */
   private void openBatch(String node)
   {
      if (!ds.openBatch(node))
      {
         throw new RuntimeException("openBatch() failed for the " + node + " node");
      }
   }
   
   /**
    * Close a batch editing session.
    * 
    * @param    node
    *           The node ID.
    * 
    * @throws   RuntimeException
    *           If the batch editing session could not be closed.
    */
   private void closeBatch(String node)
   {
      if (!ds.closeBatch(true))
         throw new RuntimeException("closeBatch() failed for " + node);
   }

   /**
    * Read a line of text using the created {@link #reader}.
    * 
    * @param    txt
    *           Description to be written to standard output.
    *           
    * @return   The line of text from standard input.
    * 
    * @throws   IOException
    *           If data could not be read.
    */
   private String readLine(String txt)
   throws IOException
   {
      System.out.print(txt);
      
      String line = reader.readLine();
   
      return line.trim();
   }

   /**
    * Create an empty store, to hold either private keys or certificates.
    * 
    * @return   An empty {@link KeyStore} instance.
    * 
    * @throws   SSLCertGenException
    *           If the store could not be generated.
    */
   private KeyStore createEmptyStore()
   throws SSLCertGenException
   {
      try
      {
         KeyStore ks = KeyStore.getInstance("JKS");
         ks.load(null, null);
         return ks;
      }
      catch (NoSuchAlgorithmException | CertificateException | IOException | KeyStoreException e)
      {
         throw new SSLCertGenException(e);
      }
   }

   /**
    * Create a random 256-bit password to be used as an AES encryption key.
    * 
    * @return   See above.
    */
   private String createAES256BitKey()
   {
      return RandomWordGenerator.create(32, 8, 4);
   }
   
   /**
    * Command line driver.
    * 
    * @param    args
    *           Application command line parameters. File name is the only one expected.
    *
    * @throws    ConfigurationException
    *            In case the {@link #ds directory service} could not be initialized.
    * @throws    IOException
    *            If standard input can not be accessed.
    * @throws    SSLCertGenException
    *            In case of problems during the generation of root CA or peer certificates.
    */
   public static void main(String[] args)
   throws ConfigurationException,
          IOException,
          SSLCertGenException
   {
      String dirFile = null;

      if (args.length == 0)
      {
         System.out.println("usage: java SSLCertGenUtil <directory.xml>");
         return;
      }
      
      dirFile = args[0];
      
      BootstrapConfig cfg = new BootstrapConfig();
      
      cfg.setConfigItem("directory", "backend", "type",        "xml");
      cfg.setConfigItem("directory", "xml",     "filename",    dirFile);
      cfg.setConfigItem("directory", "xml",     "must_exist", "true");
      cfg.setServer(true);

      SSLCertGenUtil gen = new SSLCertGenUtil(cfg);
      gen.generate();
   }
}
