/*
** Module   : ChuiWebSimulator.java
** Abstract : A WebSocket Chui Terminal simulator.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20131113 Created initial version.
** 002 MAG 20140114 Implements WebSockets protocol and communication.
** 003 MAG 20140204 Send text attributes via WebSockets. Add a Watchdog Timer 
**                  guardian and quit command.                   
** 004 CA  20140206 Web socket timeout can't be read from the jetty threads (as these don't have a
**                  P2J context, so it must be passed at construction.
*/

package com.goldencode.p2j.ui.client.chui.driver.web;

import java.io.*;
import java.nio.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;
import java.util.regex.*;

import org.eclipse.jetty.websocket.api.*;
import org.eclipse.jetty.websocket.api.annotations.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;
import com.goldencode.p2j.util.*;

/**
 * Chui WebSocket terminal simulator.
 * Implements the WebSockets protocol to communicate with the JS terminal simulator.
 * On socket close terminate the client process.
 */
@SuppressWarnings("serial")
@WebSocket(maxIdleTime = 900000, maxTextMessageSize = 4096)
public class ChuiWebSimulator
extends ChuiSimulator
{  
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ChuiWebSimulator.class.getName()); 
   
   /** Keyboard JSON pattern */
   private static final Pattern KEYBOARD = Pattern.compile("\\{\"c\":0,\"k\":([0-9]{1,4})\\}");
   
   /** Clear screen */
   private static final String INIT = "{\"c\":0}";
   
   /** Synchronization */
   private static final String SYNC = "{\"c\":1,\"t\":\"%s\"}"; 

   /** Set cursor position */
   private static final String SET_CURSOR_POS = "{\"c\":2,\"x\":\"%d\",\"y\":\"%d\"}"; 

   /** Show cursor */
   private static final String SET_CURSOR_ON = "{\"c\":3}"; 
   
   /** Hide cursor */
   private static final String SET_CURSOR_OFF = "{\"c\":4}"; 
   
   /** Play beep */
   private static final String BEEP = "{\"c\":5}"; 

   /** Quit */
   private static final String QUIT = "{\"c\":6, \"u\":\"%s\"}"; 
   
   /**  Object representing the other end of this conversation. */
   private RemoteEndpoint remote;  
   
   /** Keyboard input queue */
   private Deque<Integer> keyBuffer = new ConcurrentLinkedDeque<>();
   
   /** Keep cursor status */
   private boolean cursorVisible = true;
   
   /** Web socket timeout. */
   private final long timeout;
   
   /** Watchdog time out value */
   private final long watchdogTimeout;

   /** Watchdog instance */
   private WatchdogTimer watchdog;
   
   /**
    * Create a new chui web simulator.
    * 
    * @param    timeout
    *           Web socket timeout.
    * @param    watchdogTimeout
    *           Watchdog thread timeout.
    */
   public ChuiWebSimulator(long timeout, long watchdogTimeout)
   {
      this.timeout = timeout;
      this.watchdogTimeout = watchdogTimeout;
      
      // Start watchdog timer
      startWatchdog(watchdogTimeout);
   }
   
   /**
    * This method is called when the remote peer open a web socket connection.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    *
    */
   @OnWebSocketConnect
   public void onConnect(Session session)
   {
      // Stop watchdog timer
      stopWatchdog();
      // Get RemoteEndpoint reference.
      remote = session.getRemote();
      if (timeout > 0)
      {
         session.setIdleTimeout(timeout);
      }
      // synchronize tty
      sync();
   }

   /**
    * This method is called when a text message is received.
    *
    * @param   message
    *          The text message.
    */
   @OnWebSocketMessage
   public void onMessage(String message) {
      Matcher matcher = KEYBOARD.matcher(message);
      if (matcher.find())
      {
         String keystroke = matcher.group(1);
         keyBuffer.addLast(Integer.valueOf(keystroke));
      }                  
   }

   /**
    * This method is called when a binary message is received.
    *
    * @param   message
    *          Message content as an byte array.
    * @param   offset
    *          Message offset..
    * @param   length
    *          Message size.
    */
   @OnWebSocketMessage
   public void onMessage(byte[] message, int offset, int length)
   {
      // TODO implements binary protocol.
   }

   /**
    * This method is called when the connection is closed.
    * An END-ERROR KEY (F4) code is enqueued on key buffer queue.   
    *
    * @param   statusCode
    *          The status integer code.
    * @param   reason
    *          A string representing the reason.
    */
   @OnWebSocketClose
   public void onClose(int statusCode, String reason)
   {
      remote = null;
      // Start watchdog timer
      startWatchdog(watchdogTimeout);
      
      LOG.logp(Level.INFO,
               "ChuiWebSimulator.onClose",
               "",
               LogHelper.generate("Connection closed status=%d reason=%s", statusCode, reason));
   }
   
   /**
    * This method is called on communication errors.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    * @param   error
    *          A <code>Throwable</code> instance representing the error condition.
    */
   @OnWebSocketError
   public void onError(Session session, Throwable error)
   {
      session.close();
      remote = null;
      
      LOG.logp(Level.SEVERE,
               "ChuiWebSimulator.onError",
               "",
               LogHelper.generate("WebSocket communication error!"),
               error);      
   }

   /**
    * Start watchdog timer.
    * 
    * @param   watchdogTimeout
    *          Watchdog timeout value.
    */
   private void startWatchdog(long watchdogTimeout)
   {
      synchronized (keyBuffer)
      {
         if (watchdog == null)
         {
            watchdog = new WatchdogTimer(watchdogTimeout);
         }
      }      
   }

   /**
    * Stop watchdog timer.
    */
   private void stopWatchdog()
   {
      synchronized (keyBuffer)
      {
         if (watchdog != null)
         {
            watchdog.kill();
            // wait for thread exit
            try
            {
               watchdog.join(WatchdogTimer.WAIT_TO_DIE);
            }
            catch (InterruptedException e)
            {
               LOG.logp(Level.INFO,
                        "ChuiWebSimulator.stopWatchdog",
                        "",
                        LogHelper.generate("Watchdog interrupted!"),
                        e);      
            }
            finally
            {
               watchdog = null;
            }
         }
      }      
   }
   
   /**
    * Send a text message.
    * 
    * @param   message
    *          Text message.
    * 
    * @throws  IOException
    *          If cannot send the message.
    */   
   public void sendTextMessage(String message) throws IOException
   {
      if (remote != null)
      {
         remote.sendString(message);
      }
   }

   /**
    * Send a binary message.
    * 
    * @param   message
    *          Binary message.
    * 
    * @throws  IOException
    *          If cannot send the message.
    */   
   public void sendBinaryMessage(byte[] message) throws IOException
   {
      if (remote != null)
      {
         remote.sendBytes(ByteBuffer.wrap(message));
      }
   }
   
   /**
    * Synchronize tty.
    */
   private void sync()
   {
      triggerRepaint(true);
      setCursorStatus(cursorVisible);
      if (cursorVisible)
      {
         setCursor(getCursorX(), getCursorY());
      }
      beep();
   }
   
   /**
    * Read the next key from the FIFO key input buffer.  
    *
    * @return   The next key in the buffer or -1 if buffer is empty.
    */
   public int readKey()
   {
      return keyBuffer.isEmpty() ? -1 : keyBuffer.pollFirst();
   }
   
   
   @Override
   public void clear()
   {
      super.clear();
      try
      {
         sendTextMessage(INIT);
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.clear",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }
   }

   /**
    * The beep() should be implemented by sending a command on the remote end point.
    * The command is part of the communication protocol.
    *
    */
   public void beep()
   {
      try
      {
         sendTextMessage(BEEP);
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.beep",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }
   }
   
   @Override
   public boolean setCursorStatus(boolean on) throws IllegalStateException
   {
      super.setCursorStatus(on);
      try
      {
         cursorVisible = on;
         sendTextMessage(on ? SET_CURSOR_ON : SET_CURSOR_OFF);
         return true;
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.setCursorStatus",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }
      return false;
   }

   @Override
   public boolean setCursor(int x, int y)
   {
      super.setCursor(x, y);

      try
      {
         // Convert to JSON and send
         String payload = String.format(SET_CURSOR_POS, y, x);
         sendTextMessage(payload);
         return true;
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.setCursor",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }
      return false;
   }

   /**
    * Should be implemented by sending the screen status to the remote.
    * end point as a refresh screen operation.
    * The command is part of the communication protocol.
    *
    */
   @Override
   public void triggerRepaint(boolean full)
   {
      super.triggerRepaint(full);
      
      TerminalOptions options = getTerminalOptions();
      int rows = options.rows;
      int columns = options.columns;
      
      Cell[][] data = getScreenData();
      
      // Attributes binary message
      // first byte is an id (0 = attributes)
      ByteArrayOutputStream attr = new ByteArrayOutputStream(rows * columns);
      attr.write(0); 

      // Get screen characters
      StringBuffer buff = new StringBuffer();      
      for (int row = 0; row < rows; row++)
      {
         for (int col = 0; col < columns; col++)
         {
            Cell cell = data[row][col];
            // escape " character
            if (cell.character() == '"' || cell.character() == '\\')
            {
               buff.append('\\');
            }
            buff.append(cell.character());
            attr.write(cell.color().attribute());
         }
      }      
      
      try
      {
         // Send attributes
         byte[] attributes = attr.toByteArray(); 
         sendBinaryMessage(attributes);
         // Convert to JSON and send
         String payload = String.format(SYNC, buff.toString());
         sendTextMessage(payload);
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.triggerRepaint",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }      
   }

   /**
    * Remote application is about to end.
    * Notify the JS client about that by sending a referrer address.
    * The client should be redirected to referrer address.
    * 
    * @param   url
    *          Referrer URL
    */
   public void quit(String url)
   {
      try
      {
         String payload = String.format(QUIT, url);
         sendTextMessage(payload);
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.quit",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }
   }      
}